import time
import numpy as np
from scipy.interpolate import griddata
import logging
import traceback

log=logging.getLogger('root')

from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library import units
from xframe.library.gridLibrary import get_linspace_log2
from xframe.library.gridLibrary import uniformGrid_func

from scipy.spatial.kdtree import KDTree
from scipy.spatial import Delaunay 
from scipy.ndimage import gaussian_filter
from scipy.ndimage import spline_filter
from xframe import Multiprocessing


def create_lookup_array_2D(grid,new_grid,return_counts = False,coord_sys='cartesian'):
    '''assumes grids are in cartesian_coordinates'''
    if coord_sys == 'spherical':        
        grid = spherical_to_cartesian(grid[:])
        new_grid = spherical_to_cartesian(new_grid[:])
    if grid.shape[-1] == 3:
        grid = grid[...,:2]
    shape = grid.shape
    new_shape = new_grid.shape
    tree = KDTree(grid.reshape(-1,2))
    lookup_ids = tree.query(new_grid.reshape(-1,2))[1]
    unraveld_lookup = np.unravel_index(lookup_ids,shape[:-1])
    unraveld_lookup = np.array(tuple(part.reshape(new_shape[:-1]) for part in unraveld_lookup))
        
    if return_counts:
        counts = np.zeros(new_shape[0])
        shaped_lookup = lookup_ids.reshape(new_shape[:-1])
        for n,x in enumerate(shaped_lookup):
            counts[n] = len(np.unique(x))
        return unraveld_lookup,counts
    else:
        return unraveld_lookup
    

def get_min_diffs(array):
    min_diffs = []
    for dim in range(array.ndim-1):
        min_diff = np.abs(np.diff(array[...,dim],axis = dim)).min()
        min_diffs.append(min_diff)
    return np.array(min_diffs)

def get_max_diffs(array):
    max_diffs = []
    for dim in range(array.ndim-1):
        max_diff = np.abs(np.diff(array[...,dim],axis = dim)).max()
        max_diffs.append(max_diff)
    return np.array(max_diffs)



class SimpleRegridder2D:
    def __init__(self,pixel_centers,new_grid,interpolation = 'nearest',coord_sys='spherical',interpolation_constants=False):
        ''' assumes pixel_centers in spherical coordinates'''
        self.out_shape = new_grid.shape[:-1]
        self.in_shape = pixel_centers.shape[:-1]
        self.pixel_centers = pixel_centers
        self.new_grid = new_grid
        self.interpolation = interpolation
        self.interpolation_constants = interpolation_constants
        if coord_sys == 'spherical':
            self.pixel_centers = spherical_to_cartesian(pixel_centers)
            self.new_grid = spherical_to_cartesian(new_grid)
        self.pixel_centers_hash,self.new_grid_hash = self.generate_hashes(self.pixel_centers,self.new_grid)
        
        try:
            self.interpolation_constants = self.generate_interpolation_constants()
        except AttributeError as e:
            traceback.print_exc()
            log.error('interpolation method {} not known. Available interpolation options are {}'.format(interpolation,'"nearest","linear"')) 
        self.apply = getattr(self,'generate_apply_'+interpolation)()
        
    def generate_hashes(self,pixel_centers,new_grid):
        pixel_centers_hash = hash(tuple(pixel_centers.reshape(-1)[:1000]))
        new_grid_hash = hash(tuple(new_grid.reshape(-1)[:1000]))
        return pixel_centers_hash,new_grid_hash
    def check_interpolation_constants(self,interpolation_constants):
        pixel_centers_match = self.pixel_centers_hash == interpolation_constants['hash_pixel_centers']
        new_grids_match = self.new_grid_hash == interpolation_constants['hash_new_grid']

        new_interpolation = interpolation_constants['interpolation']
        if isinstance(new_interpolation,bytes):
            new_interpolation = new_interpolation.decode()
        interpolation_match = self.interpolation == new_interpolation
        constants_valid = (new_grids_match and pixel_centers_match and interpolation_match)
        #log.info('constants valid {}'.format(constants_valid))
        #log.info('new_grid hash {} {}'.format(self.new_grid_hash,interpolation_constants['hash_new_grid']))
        #log.info('pixel_centers hash {} {}'.format(self.pixel_centers_hash,interpolation_constants['hash_pixel_centers']))
        return constants_valid
    
    def assemble_interpolation_dict(self, data: dict):
        interpolation_constants={
            'hash_pixel_centers': self.pixel_centers_hash,
            'hash_new_grid': self.new_grid_hash,
            'interpolation':self.interpolation,
            **data
        }
        return interpolation_constants

    def generate_interpolation_constants(self):
        constants_fit_to_grids = False
        interpolation_constants = self.interpolation_constants
        interpolation = self.interpolation
        if isinstance(interpolation_constants,dict):
            constants_fit_to_grids = self.check_interpolation_constants(interpolation_constants)
        if not constants_fit_to_grids:
            try:
                interpolation_constants = getattr(self,'generate_interpolation_constants_'+interpolation)(interpolation_constants)
            except AttributeError as e:
                traceback.print_exc()
                log.error('interpolation method {} not known. Available interpolation options are {}'.format(interpolation,'"nearest","linear"'))
        return interpolation_constants
            
    def generate_interpolation_constants_nearest(self,interpolation_constants):
        constants_fit_to_grids = False
        if isinstance(interpolation_constants,dict):
            constants_fit_to_grids = self.check_interpolation_constants(interpolation_constants)
        if not constants_fit_to_grids:
            lookup_array,counts = create_lookup_array_2D(self.pixel_centers,self.new_grid,return_counts = True)
            data = {'lookup_array':lookup_array}
            interpolation_constants = self.assemble_interpolation_dict(data)
        return interpolation_constants    
    def generate_interpolation_constants_linear(self,interpolation_constants):
        constants_fit_to_grids = False
        #log.info('inerpolation constants type is {}'.format(type(interpolation_constants)))
        if isinstance(interpolation_constants,dict):
            constants_fit_to_grids = self.check_interpolation_constants(interpolation_constants)
        if not constants_fit_to_grids:
            vertices,weights = self.generate_linear_interp_weights(self.pixel_centers,self.new_grid)            
            data = {'vertices':vertices,'weights':weights}
            interpolation_constants = self.assemble_interpolation_dict(data)            
        return interpolation_constants

    def generate_linear_interp_weights(self,grid,new_grid):
        d = grid.shape[-1]
        g = grid.reshape(-1,d)
        ng = new_grid.reshape(-1,d)
        #log.info('grid shape = {} new grid shape = {}'.format(g.shape,ng.shape))
        tri = Delaunay(g)
        simplex = tri.find_simplex(ng)
        outside_hull = (simplex == -1)
        vertices = tri.simplices[simplex]
        temp = tri.transform[simplex]
        delta = ng - temp[:, d]
        bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
        weights = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))
        weights[outside_hull] = 0
        return vertices, weights

    
    def generate_apply_linear(self):
        weights = self.interpolation_constants['weights']
        vertices = self.interpolation_constants['vertices']
        shape = self.out_shape
        ndim_in = len(self.in_shape)
        def apply(data):
            ndim = data.ndim
            N=data.shape[-1]
            if ndim >= 1+ndim_in:
                invariant_shape = data.shape[:-2]
                N_invar = np.prod(invariant_shape)
                data = np.moveaxis(np.moveaxis(data,-1,0),-1,0)
                return np.einsum('njk,nj->nk', data.reshape(-1,N_invar)[vertices], weights).T.reshape(invariant_shape+shape)
            else:                
                return np.einsum('nj,nj->n', data.reshape(-1)[vertices], weights).reshape(shape)
        return apply
    
    def generate_apply_nearest(self):
        lookup = self.interpolation_constants['lookup_array'].astype(int)
        def apply(data):
            if data.ndim == 1+len(lookup):
                return data[tuple(lookup)]
            else:
                return data[tuple(lookup)]
        return apply

class AgipdRegridderSimple:
    # Regridder for AGIPD data in the 16,512,128 format.
    # Uses a SimpleRegridder2D object for each of the 16 modules separatelly.    
    version = 0.3
    @staticmethod
    def get_interpolation_constants_name(geometry,new_grid_shape,interpolation):

        pixel_hash = hash(tuple(geometry['framed_pixel_centers'].reshape(-1)[:1000]))
        #log.info('new_grid shape = {}'.format(new_grid_shape))
        new_grid_hash = hash(new_grid_shape + (AgipdRegridderSimple.version,))
        return '{}_{}_{}'.format(interpolation,pixel_hash,new_grid_hash)
    
    
    def __init__(self,geometry,new_grid_shape, interpolation = 'nearest', interpolation_constants=False):        
        spherical_pixel_centers = geometry['framed_pixel_centers']
        self.default_interpolation_constants = {
            'data':{str(m_id):False for m_id in range(len(spherical_pixel_centers))},
            'mask':{str(m_id):False for m_id in range(len(spherical_pixel_centers))},
            'module_masks':np.array([False]*len(spherical_pixel_centers)),
            'hash_name': 'not assigned'
        }
        #self.pixel_centers = geometry['framed_pixel_centers']
        self.geometry = geometry
        self.sensitve_pixel_mask = geometry['framed_mask']
        self.interpolation = interpolation
        self.new_grid_shape = new_grid_shape
        self.max_pixel_q = np.max(spherical_pixel_centers[...,0])
        self.polar_pixel_centers = spherical_pixel_centers[...,::2]
        self.cart_pixel_centers = spherical_to_cartesian(self.polar_pixel_centers)
        #self.polar_pixel_centers = cartesian_to_spherical(self.cart_pixel_centers)
        #cart_pixel_centers = spherical_to_cartesian(polar_pixel_centers)
        
        self.new_grid = self._generate_new_grid(new_grid_shape,self.max_pixel_q)
        self.new_qs = self.new_grid[:,0,0]
        self.new_phis = self.new_grid[0,:,1]

        self.generate_new_constants,self.interpolation_constants = self.process_interpolation_constants(interpolation_constants)
        if self.generate_new_constants:
            self.module_masks = self._generate_module_masks()
            self.interpolation_constants['module_masks']=self.module_masks
            temp = self._generate_regridders()
            self.data_regridders_per_module = temp[0]
            self.mask_regridders_per_module = temp[1]
            self.interpolation_constants = self._assemble_interpolation_constants()
        else:            
            temp = self._generate_regridders()
            self.data_regridders_per_module = temp[0]
            self.mask_regridders_per_module = temp[1]
        self.regrid = self.generate_regrid()

    def _generate_new_grid(self,new_grid_shape,max_q):
        Nq,Nphi = new_grid_shape
        qs = max_q*np.arange(Nq)/(Nq-1)
        phis = 2*np.pi*np.arange(Nphi)/Nphi
        new_grid = GridFactory.construct_grid('uniform',[qs,phis])[:]
        return new_grid

    def _generate_module_masks(self):
        def delaunay_mask_finding(module_ids,rough_module_masks,cart_new_grid,in_masks,**kwargs):
            module_masks = [False]*len(module_ids)
            
            #log.info(f'module_ids = {module_ids}')
            #log.info(f'module_masks shape  = {rough_module_masks.shape}')
            #log.info(f'cart_new_grid shape  = {cart_new_grid.shape}')
            for _id,mid in enumerate(module_ids):
                centers = self.cart_pixel_centers[mid]
                #delaunay = Delaunay(centers.reshape(-1,2))
                rough_module_mask = rough_module_masks[mid]
                in_mask = in_masks[mid]
                new_points = cart_new_grid[rough_module_mask]
                regridder = SimpleRegridder2D(centers,new_points,coord_sys='cartesian')
                mask = rough_module_mask.copy()
                mask[mask] = regridder.apply(in_mask).astype(bool)                
                module_masks[_id] = mask                
            return np.asarray(module_masks)
        new_grid = self.new_grid
        in_masks = np.ones(self.sensitve_pixel_mask.shape,dtype = bool)
        in_masks[:,0,:]=0
        in_masks[:,-1,:]=0
        in_masks[:,:,0]=0
        in_masks[:,:,-1]=0
        # calculate rough module boundary box
        module_boundaries = []
        for dim in range(2):
            _min = np.min(self.cart_pixel_centers[:,...,dim],axis = (1,2))
            _max = np.max(self.cart_pixel_centers[:,...,dim],axis = (1,2))
            #log.info(f'dim {dim} : min {_min} max {_max} ')
            module_boundaries.append([_min,_max])        
        module_boundaries = np.moveaxis(np.array(module_boundaries),-1,0)        
        n_modules = module_boundaries.shape[0]

        # improve module masks by delaunay triangulation.
        # This is to avoid overlapping module masks.
        cart_grid = spherical_to_cartesian(new_grid)
        min_mask = cart_grid>=module_boundaries[:,None,None,:,0]
        max_mask = cart_grid<=module_boundaries[:,None,None,:,1]
        temp_module_masks = min_mask[...,0] & min_mask[...,1] & max_mask[...,0] & max_mask[...,1]
        module_masks = Multiprocessing.comm_module.request_mp_evaluation(delaunay_mask_finding,input_arrays=[np.arange(n_modules)],const_inputs = [temp_module_masks,cart_grid,in_masks] ,call_with_multiple_arguments = True,split_mode='modulus',n_processes= False)
        #log.info(f'module_masks .shape = {module_masks.shape} | {module_masks.dtype}')
        return module_masks
    
    def _generate_regridders(self):
        new_grid = self.new_grid
        polar_pixel_centers = self.polar_pixel_centers
        constants = self.interpolation_constants
        module_masks = constants['module_masks']
        #log.info(f'masks .shape = {module_masks.shape}')
        data_regridders = {}
        mask_regridders = {}        
        for m_id,mask in enumerate(module_masks):
            module_pixel_centers = polar_pixel_centers[m_id]
            new_points = new_grid[mask]
                
            r = SimpleRegridder2D(module_pixel_centers,new_points,interpolation = self.interpolation,interpolation_constants = constants['data'][str(m_id)])
            if self.interpolation == 'linear':
                rm = SimpleRegridder2D(module_pixel_centers,new_points,interpolation = 'nearest',interpolation_constants = constants['mask'][str(m_id)])
            else:
                rm = r
            data_regridders[str(m_id)] = r
            mask_regridders[str(m_id)] = rm        
        return data_regridders,mask_regridders
    def _assemble_interpolation_constants(self):
        data_constants = {str(m_id):r.interpolation_constants for m_id,r in self.data_regridders_per_module.items()}
        mask_constants = {str(m_id):r.interpolation_constants for m_id,r in self.mask_regridders_per_module.items()}
        hash_name = self.get_interpolation_constants_name(self.geometry,self.new_grid_shape,self.interpolation)        
        interpolation_constants={'data' : data_constants,'mask':mask_constants,'hash_name':hash_name,'module_masks':self.module_masks}
        return interpolation_constants
    def process_interpolation_constants(self,constants):
        constants_are_valid = False
        if isinstance(constants,dict):            
            hash_name = self.get_interpolation_constants_name(self.geometry,self.new_grid_shape,self.interpolation)                
            constants_name = constants['hash_name']
            if isinstance(constants_name,bytes):
                constants_name = constants_name.decode()
            constants_are_valid = (hash_name == constants_name)
            #log.info('hash_names = {} {}'.format(hash_name,constants_name))
        if constants_are_valid:
            constants['module_masks'] = np.asarray(constants['module_masks'],dtype = bool) 
            interpolation_constants = constants
        else:
            interpolation_constants = self.default_interpolation_constants
        return (not constants_are_valid),interpolation_constants
    def generate_regrid(self):
        in_mask = self.sensitve_pixel_mask
        in_shape = self.sensitve_pixel_mask.shape
        out_shape = self.new_grid_shape
        module_masks = self.interpolation_constants['module_masks']
        data_regridders = self.data_regridders_per_module
        mask_regridders = self.mask_regridders_per_module
        new_masks = [1]*len(module_masks)
        def regrid(data,mask,modules,**args):
            ndim = data.ndim
            #log.info('regrid ndim ={}'.format(ndim))
            #log.info('in mask shape ={}'.format(in_mask.shape))
            if ndim == in_mask.ndim+1:
                N = len(data)
                # have to copy data to framed agipd shape
                data_dtype = data.dtype
                mask_dtype = mask.dtype
                in_d = np.zeros(in_shape+(N,),data_dtype)
                in_m = np.zeros(in_shape+(N,),mask_dtype)
                out_d = np.zeros((N,)+out_shape,data_dtype)
                out_m = np.zeros((N,)+out_shape,mask_dtype)
                                
                in_d[in_mask] = np.moveaxis(data,0,-1).reshape(-1,N)
                in_m[in_mask]= np.moveaxis(mask,0,-1).reshape(-1,N)
                for m in modules:
                    mask = module_masks[m]
                    out_d[:,mask]= np.moveaxis(data_regridders[str(m)].apply(in_d[m]),-1,0)
                    out_m[:,mask]= np.moveaxis(mask_regridders[str(m)].apply(in_m[m]),-1,0)                
            else:
                data_dtype = data.dtype
                mask_dtype = mask.dtype
                in_d = np.zeros(in_shape,data_dtype)
                in_m = np.zeros(in_shape,mask_dtype)
                out_d = np.zeros(out_shape,data_dtype)
                out_m = np.zeros(out_shape,mask_dtype)
                in_d[in_mask] = data.reshape(-1)
                in_m[in_mask]= mask.reshape(-1)
                
                for m in modules:                    
                    mask = module_masks[m]
                    #log.info(f'mask shape = {mask.shape}')
                    #log.info(f'masks shape = {module_masks.shape}')
                    out_d[mask]= data_regridders[str(m)].apply(in_d[m])
                    out_m[mask]= mask_regridders[str(m)].apply(in_m[m])
            return out_d,out_m
        return regrid
                

### old stuff ###
class NearestNeighbourAgipdRegridder:
    def __init__(self,pixel_centers,polar_grid):
        ''' assumes pixel_centers in spherical coordinates'''
        self.cart_pixel_centers = spherical_to_cartesian(pixel_centers)
        self.cart_grid = spherical_to_cartesian(self.polar_grid)
        self.lookup_array,self.counts = create_lookup_array_2D(self.cart_pixel_centers,self.cart_grid,return_counts = True)
        self.apply = self.generate_apply()
    def generate_apply(self):
        lookup = self.lookup_array
        def apply(data):
            return data[lookup]
        return apply
    def generate_regrid_with_gaussian(self,intermediate_sampling=1,sigma=1):
        '''
        input grid -> intermediate uniform grid ---[gaussian_filter]---> intermediate uniform grid -> output grid  
        intermediate_sampling defines the step size of the intermediate uniform grid in terms of multiples of the minimalpoint distance in the input grid.
        sigma defines the standard deviation of the gaussian filter in units of the maximal point distance in the inputgrid.
        Assumes pixel centers to be of shape [modules,x,y,point]
        '''
        assert intermediate_sampling>=1 , 'intermediate_sampling has to be greater than 1'
        min_diff = np.min([get_min_diffs(m) for m in self.cart_pixel_centers])/intermediate_sampling
        max_diff = np.max([get_max_diffs(m) for m in self.cart_pixel_centers],axis = 0)
        max_diff = np.max(max_diff[0]/2,max_diff[1]) # accounting for the dead wide pixels in AGIPD        
        bounds = np.max(cart_pixel_centers,axis = (0,1,2))[:2]
        n_steps = (bounds//min_diff + 1).astype(int)
        xs = np.arange(n_teps[0])*bounds[0]/n_steps[0]
        ys = np.arange(n_teps[1])*bounds[1]/n_steps[1]
        intermediate_grid = GridFactory('uniform',[xs,ys])
        lookup_inter = create_lookup_array_2D(self.cart_pix_centers,intermediate_grid)
        lookup_polar = create_lookup_array_2D(intermediate_grid,self.polar_grid)
        sigma = max_diff*sigma

        def apply_with_gaussian(data):
            intermediate_data = data[lookup_inter]
            intermediate_data = gaussian_filter(intermediate_data,sigma = sigma)
            return intermediate_data[lookup_polar]
        return apply_with_gaussian
                
class AgipdRegridder:
    def __init__(self,db,geometry,wavelength, file_name='regrid_data', save_regrid_data=False, generate_data=False):
        #framed_pixel_grid in spherical_coordinates
        self.pixel_grid_spher = geometry['framed_pixel_grid']
        self.pixel_grid_spher[...,0]*=units.standardLength
        self.lab_pixel_grid_cart = geometry['framed_lab_pixel_grid']
        self.pixel_mask = geometry['framed_mask']
        self.data_shape = geometry['data_shape']
        self.wavelength = wavelength
        self.file_name = file_name
        self.db=db
        if generate_data:
            self.regrid_data=self.generate_regrid_data()
            if save_regrid_data:
                db.save(file_name,self.regrid_data)
        else:
            try:
                self.regrid_data=db.load(file_name)
            except FileNotFoundError as e:
                log.warning('regrid_data file not found. Proceed by generating regrid_data.')
                self.regrid_data=self.generate_regrid_data()
                if save_regrid_data:
                    db.save(file_name,self.regrid_data)

    def approximate_pixel_size(self,grid_cart):
        #absolute minimal x-distance between consecutive agipd pixels along vertical(short direction 128 pixels per module) pixel row
        x_min = np.abs([np.diff(grid_cart[m][:,:,0],axis=1).min() for m in np.arange(len(grid_cart))]).min()
        #same but in y-direction
        y_min = np.abs([np.diff(grid_cart[m][:,:,1],axis=1).min() for m in np.arange(len(grid_cart))]).min()
        #same but in z-direction
        z_min = np.abs([np.diff(grid_cart[m][:,:,2],axis=1).min() for m in np.arange(len(grid_cart))]).min()        
        approximated_pixel_size= np.max([x_min,y_min,z_min])
        log.info('approximate_pixel_size = {}'.format(approximated_pixel_size))
        return approximated_pixel_size


    def generate_regrid_data(self):
        pixel_grid_cart = spherical_to_cartesian(self.pixel_grid_spher)
        data = self.generate_regrid_data_polar(self.pixel_grid_spher,grid_point_multiplicator = [3,1])    
        data2 = self.generate_regrid_data_cart(self.lab_pixel_grid_cart,grid_point_multiplicator = [4,4])
        lab_small = self.generate_regrid_data_cart(self.lab_pixel_grid_cart,grid_point_multiplicator = [2,2])
        q_cart_small = self.generate_regrid_data_cart(pixel_grid_cart,grid_point_multiplicator = [2,2])
        
        cart_data = {'lab':data2,'lab_small':lab_small,'q_small':q_cart_small}
        polar_data = {'reciprocal':data}
        regrid_data = {'cart':cart_data,'polar':polar_data}
        return regrid_data


    def generate_regrid_data_cart(self,pixel_grid,grid_point_multiplicator=[4,4],on_q_space=False):
        regrid_data={}
        d_shape = self.data_shape
        
        pix_size=self.approximate_pixel_size(pixel_grid)        
        
        old_global_x_bounds = (pixel_grid[...,0].min(),pixel_grid[...,0].max())
        old_global_y_bounds = (pixel_grid[...,1].min(),pixel_grid[...,1].max())

        try:
            assert (old_global_x_bounds[0]<=0) and (old_global_x_bounds[1]>0), 'global x bounds: ( {} ) do not contain 0.'.format(old_global_x_bounds)
            assert (old_global_y_bounds[0]<=0) and (old_global_y_bounds[1]>0), 'global y bounds: ( {} ) do not contain 0.'.format(old_global_y_bounds)
        except AssertionError as e:
            log.error(e)
            traceback.print_exc()
            raise e

        target_x_step = pix_size/grid_point_multiplicator[0]
        xs,x_param = get_linspace_log2(old_global_x_bounds,target_x_step,additional_steps = 1,return_parameters = True)
        x_step = x_param['step_size']
        global_x_bounds = x_param['domain']

        target_y_step = pix_size/grid_point_multiplicator[1]
        ys,y_param = get_linspace_log2(old_global_y_bounds,target_y_step,additional_steps = 1, return_parameters = True)
        y_step = y_param['step_size']
        global_y_bounds = y_param['domain']

        out_grid = GridFactory.construct_grid('uniform',[xs,ys]).array
        
        regrid_data['x_step'] = x_step
        regrid_data['xs'] = xs
        regrid_data['y_step'] = y_step
        regrid_data['ys'] = ys
        regrid_data['grid'] = out_grid

        global_x_min = global_x_bounds[0]
        global_y_min = global_y_bounds[0]
        log.info('out grid shape = {}'.format(out_grid.shape))
        for m,g in enumerate(pixel_grid):
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]
        
            x_bounds = [g[...,0].min(),g[...,0].max()]            
            x_min_index = max(0,int((x_bounds[0]-global_x_min)//x_step)+1)
            x_max_index = min(int((x_bounds[1]-global_x_min)//x_step)+1,out_grid.shape[0])        

            y_bounds = [g[...,1].min(),g[...,1].max()]
            #log.info('x_bounds = {}'.format(x_bounds))
            y_min_index = max(int((y_bounds[0]-global_y_min)//y_step)+1,0)
            y_max_index = min(int((y_bounds[1]-global_y_min)//y_step)+1,out_grid.shape[1])        
            m_grid = out_grid[x_min_index:x_max_index,y_min_index:y_max_index]
            log.info("x_range = {} y_range ={}".format([x_max_index,x_min_index],[y_max_index,y_min_index]))
            #log.info("x_range = {} y_range ={}".format([x_max_index-x_min_index],[y_max_index-y_min_index]))
            log.info('mgrid.shape = {}'.format(m_grid.shape))
            old_m_grid =g[...,:2]
            old_index = np.zeros(g.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
        
            index = griddata(old_m_grid.reshape(-1,2),old_index.flatten(),m_grid.reshape(-1,2),method='nearest',rescale = True).reshape(m_grid.shape[:-1])
            #log.info('index.shape = {}'.format(index.shape))
            #log.info('ranges y: {} x:{}'.format(y_max_index-y_min_index,x_max_index - x_min_index))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['x_range']=[x_min_index,x_max_index]
            m_dict['y_range']=[y_min_index,y_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data

    def generate_regrid_data_polar(self,pixel_grid,grid_point_multiplicator=[3,1],on_q_space=False):
        regrid_data={}
        d_shape = self.data_shape

        pixel_grid_c = spherical_to_cartesian(pixel_grid)
        pix_size=self.approximate_pixel_size(pixel_grid_c)
        
        old_global_r_bounds = (pixel_grid[...,0].min(),pixel_grid[...,0].max())
        old_global_phi_bounds = (pixel_grid[...,2].min(),pixel_grid[...,2].max())
        log.info('phi bounds = {}'.format(old_global_phi_bounds))

        target_r_step = pix_size/grid_point_multiplicator[0]
        r_length = old_global_r_bounds[1]-old_global_r_bounds[0]
        n_r_steps = int(r_length//target_r_step)
        r_step = r_length/(n_r_steps-1)
        rs = old_global_r_bounds[0]+ np.arange(n_r_steps)*r_step
        global_r_bounds = old_global_r_bounds
        log.info('r bounds = {} max,min = {}'.format(global_r_bounds,(rs.min(),rs.max())))


        # phi_step such that the arc length at max_x is proportional to the approximated pixel size    
        target_phi_step = pix_size/(global_r_bounds[1]*grid_point_multiplicator[1])
        phis,phi_param = get_linspace_log2(old_global_phi_bounds,target_phi_step,additional_steps = 1, return_parameters = True)
        phi_step = phi_param['step_size']
        global_phi_bounds = phi_param['domain']
        log.info('phi bounds = {} max,min = {}'.format(global_phi_bounds,(phis.min(),phis.max())))
                
        thetas = ewald_sphere_theta(self.wavelength,rs)
        tmp_grid = GridFactory.construct_grid('uniform',[rs,phis]).array
        out_grid = np.zeros(tmp_grid.shape[:-1]+(tmp_grid.shape[-1]+1,))
        out_grid[...,[0,2]] = tmp_grid
        out_grid[...,1] = thetas[:,None]
        
        
        regrid_data['r_step'] = r_step
        regrid_data['rs'] = rs
        regrid_data['phi_step'] = phi_step
        regrid_data['phis'] = phis
        regrid_data['grid'] = out_grid

        global_r_min = global_r_bounds[0]
        global_phi_min = global_phi_bounds[0]
        log.info('out grid shape = {}'.format(out_grid.shape))
        
        out_grid_c = spherical_to_cartesian(out_grid)[...,:2]
        pixel_grid_c = spherical_to_cartesian(pixel_grid)[...,:2]
        for m,g in enumerate(pixel_grid):            
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]

            g_c = pixel_grid_c[m]
        
            r_bounds = [g[...,0].min(),g[...,0].max()]            
            r_min_index = max(int((r_bounds[0]-global_r_min)//r_step)+1,0)
            r_max_index = min(int((r_bounds[1]-global_r_min)//r_step)+1,out_grid.shape[0])        

            phi_bounds = [g[...,2].min(),g[...,2].max()]
            #log.info('global phi_min = {}'.format(global_phi_min))
            #log.info('phi_bounds = {}'.format(phi_bounds))
            phi_min_index = max(int((phi_bounds[0]-global_phi_min)//phi_step)+1,0)
            phi_max_index = min(int((phi_bounds[1]-global_phi_min)//phi_step)+1,out_grid.shape[1])     
            out_g_c = out_grid_c[r_min_index:r_max_index,phi_min_index:phi_max_index]
            #log.info('min - glob =  {}'.format((phi_bounds[0]-global_phi_min)//phi_step+1))
            #log.info('max - glob =  {}'.format((phi_bounds[1]-global_phi_min)//phi_step+1))
            #log.info('out_grid shape = {}'.format(out_grid.shape))
            #log.info("r_range = {} phi_range ={}".format([r_max_index,r_min_index],[phi_max_index,phi_min_index]))
            #log.info("x_range = {} y_range ={}".format([x_max_index-x_min_index],[y_max_index-y_min_index]))
            log.info('mgrid.shape = {}'.format(out_g_c.shape))
            old_index = np.zeros(g.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
        
            index = griddata(g_c.reshape(-1,2),old_index.flatten(),out_g_c.reshape(-1,2),method='nearest',rescale = True).reshape(out_g_c.shape[:-1])
            #log.info('index nonzero: {}'.format((index!=0).any()))
            #log.info('index.shape = {}'.format(index.shape))
            #log.info('ranges y: {} x:{}'.format(y_max_index-y_min_index,x_max_index - x_min_index))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['r_range']=[r_min_index,r_max_index]
            m_dict['phi_range']=[phi_min_index,phi_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data    
    
    def generate_regrid_data_polar_old(self,pixel_grid,grid_point_multiplicator=[2,1]):
        regrid_data={}
        d_shape = self.data_shape

        grid_c = spherical_to_cartesian(pixel_grid)
        pix_size=self.approximate_pixel_size(grid_c)
        
        r_max = pixel_grid[...,0].max()
        # r_step such that it is proportional to the approximated pixel size
        target_r_step = (pix_size/grid_point_multiplicator[0])
        n_new_rs = (r_max//target_r_step)
        r_step = r_max/n_new_rs
        rs= np.arange(n_new_rs)*r_step

        thetas = ewald_sphere_theta(self.wavelength,rs)

        # phi_step such that the arc length at max_x is proportional to the approximated pixel size
        
        target_phi_step = pix_size/(r_max*grid_point_multiplicator[1])
        phis, phi_param = get_linspace_log2([0,2*np.pi -target_phi_step],target_phi_step,additional_steps = 1,return_parameters = True)
        phi_step = phi_param['step_size']

        tmp_grid = GridFactory.construct_grid('uniform',[rs,phis]).array
        out_grid = np.zeros(tmp_grid.shape[:-1]+(tmp_grid.shape[-1]+1,))
        out_grid[...,[0,2]] = tmp_grid
        out_grid[...,1] = thetas[:,None]
        
        regrid_data['r_step'] = r_step
        regrid_data['rs'] = rs
        regrid_data['phi_step'] = phi_step
        regrid_data['phis'] = phis
        regrid_data['grid'] = out_grid
        log.info('out grid shape = {}'.format(out_grid.shape))
        log.info('out grid shape = {}'.format(out_grid.shape))
        grid_c = spherical_to_cartesian(out_grid)[...,:2]
        pixel_grid_c = spherical_to_cartesian(pixel_grid)[...,:2]
        
        for m,g in enumerate(pixel_grid):
            log.info('generating regrid_data for module {}/{}'.format(m,len(pixel_grid)))
            #log.info('g shape = {}'.format(g.shape))
            g_c=pixel_grid_c[m]
            #log.info('g_c shape = {}'.format(g_c.shape))
            regrid_data[str(m)]={}
            m_dict=regrid_data[str(m)]
        
            r_bounds = [g[...,0].min(),g[...,0].max()]
            r_min_index = int(r_bounds[0]//r_step)
            r_max_index = int(r_bounds[1]//r_step)        
            
            phi_bounds = [g[...,2].min(),g[...,2].max()]
            #log.info('phi_bounds = {}'.format(phi_bounds))
            phi_min_index = int((np.pi+phi_bounds[0])//phi_step)
            phi_max_index = int((np.pi+phi_bounds[1])//phi_step)        
            m_grid = grid_c[r_min_index:r_max_index,phi_min_index:phi_max_index]
            log.info("r_range = {} phi_range ={}".format([r_min_index,r_max_index],[phi_min_index,phi_max_index]))
            #log.info('mgrid.shape = {}'.format(m_grid.shape))
            old_m_grid =g_c
            old_index = np.zeros(g_c.shape[:-1])
            old_index[:-1,:-1][self.pixel_mask[m]] = np.arange(1,d_shape[1]*d_shape[2]+1)
            log.info('old index nonzero = {}'.format((old_index !=0).any()))
            #log.info('old m grid  = {}'.format(old_m_grid.shape))
            
            index = griddata(old_m_grid.reshape(-1,2),old_index.flatten(),m_grid.reshape(-1,2),method='nearest',rescale = True).reshape(m_grid.shape[:-1])
            log.info('index nonzero = {}'.format((index!=0).any()))
            #log.info('ranges r: {} phi:{}'.format(r_max_index-r_min_index,phi_max_index - phi_min_index))
            #log.info('index.shape = {}'.format(index.shape))
            #new_data = griddata(old_grid,old_data.flatten(),new_grid,method='linear')

            m_dict['index']=index
            m_dict['r_range']=[r_min_index,r_max_index]
            m_dict['phi_range']=[phi_min_index,phi_max_index]
            valid_index_mask = (index!=0)
            m_dict['valid_data_mask'] = valid_index_mask
            m_dict['valid_index'] = (index[valid_index_mask]-1).astype(int)            
        return regrid_data
    
    def generate_and_save_regrid_data(self):
        self.regrid_data=self.generate_regrid_data()
        self.db.save(self.file_name,self.regrid_data)

        

    def put_data_into_grid(self,data,out_grid,r_range,phi_range):
        new_data=np.zeros((data.shape[0],)+out_grid.shape[:-1])
        #log.info('new data shape = {}'.format(new_data.shape))
        #log.info("r_range = {} phi_range ={}".format(r_range,phi_range))
        #log.info("len r = {} len phi ={} data shape = {}".format(np.diff(r_range),np.diff(phi_range),data.shape))
        new_data[:,r_range[0]:r_range[1],phi_range[0]:phi_range[1]] = data
        return new_data
        
    def regrid_polar(self,data, combine_modules=False, modules_on_full_grid=False,grid_type='reciprocal'):
        regrid_polar_dict = self.regrid_data['polar']
        for known_grid_type in regrid_polar_dict:
            if grid_type == known_grid_type:
                regrid_data = regrid_polar_dict[known_grid_type]
                break
            
        out_grid = regrid_data['grid']
        out_grid_max_index = regrid_data['grid']
        
        out_data=[]
        for m in np.arange(len(data)):
            #log.info("m = {} of {}".format(m,len(data)))
            valid_data_mask=regrid_data[str(m)]['valid_data_mask'].astype(bool)
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            valid_index=regrid_data[str(m)]['valid_index']
            #log.info('valid_index = {}'.format(valid_index.shape))

            if len(data.shape) == len(self.data_shape):
                data = data[None,...]
            regridded_m_data=np.zeros((data.shape[0],)+valid_data_mask.shape)
            regridded_m_data[:,valid_data_mask]=data[:,m].reshape(data.shape[0],-1)[:,valid_index]
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            if (combine_modules or modules_on_full_grid):
                r_range = regrid_data[str(m)]['r_range']
                phi_range = regrid_data[str(m)]['phi_range']
                regridded_m_data = self.put_data_into_grid(regridded_m_data,out_grid,r_range,phi_range)[:,:-1,:-1]
            out_data.append(regridded_m_data)            
            
        if combine_modules:
            out_data = np.sum(out_data,axis=0)
        if modules_on_full_grid:
            out_data = np.array(out_data)
        if isinstance(out_data,np.ndarray):
            out_data = np.squeeze(out_data)
            
        return out_data
    
    def regrid_cart(self,data, combine_modules=False, modules_on_full_grid=False,grid_type='lab'):
        regrid_cart_dict = self.regrid_data['cart']
        #log.info(regrid_cart_dict.keys())
        for known_grid_type in regrid_cart_dict:
            if grid_type == known_grid_type:
                regrid_data = regrid_cart_dict[known_grid_type]
                break
        
        out_grid = regrid_data['grid']
        out_data=[]
        for m in np.arange(len(data)):
            #log.info("m = {} of {}".format(m,len(data)))
            valid_data_mask=regrid_data[str(m)]['valid_data_mask'].astype(bool)
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            valid_index=regrid_data[str(m)]['valid_index']
            #log.info('valid_index = {}'.format(valid_index.shape))
            #log.info('valid_data_mask shape = {}'.format(valid_data_mask.shape))
            #log.info('regridded_m_data shape = {}'.format(regridded_m_data[~valid_data_mask].shape))
            if len(data.shape) == len(self.data_shape):
                data = data[None,...]
            regridded_m_data=np.zeros((data.shape[0],)+valid_data_mask.shape)
            regridded_m_data[:,valid_data_mask]=data[:,m].reshape(data.shape[0],-1)[:,valid_index]
            #log.info("regridded shape = {}".format(regridded_m_data.shape))
            
            if (combine_modules or modules_on_full_grid):
                r_range = regrid_data[str(m)]['x_range']
                phi_range = regrid_data[str(m)]['y_range']
                #log.info("x_range: {} y_range: {}".format(r_range,phi_range))
                regridded_m_data = self.put_data_into_grid(regridded_m_data,out_grid,r_range,phi_range)[:,:-1,:-1]                
                
            out_data.append(regridded_m_data)            
            
        if combine_modules:
            out_data = np.sum(out_data,axis=0)
        if modules_on_full_grid:
            out_data = np.array(out_data)
        if isinstance(out_data,np.ndarray):
            out_data = np.squeeze(out_data)
            
        return out_data



import numpy as np
from xframe.externalLibraries import soft_plugin,shtns_plugin

def test_rotate_coeff():
    bw=4
    shtns = shtns_plugin.sh(bw-1)
    soft = soft_plugin.Soft(bw)
    
    coeff = np.zeros(shtns.n_coeff,np.complex128)
    #coeff[4:9]=1
    coeff[4:9]=np.array([1.1, 0, 2, 3, 1])
    split_ids = shtns.cplx_l_split_indices
    euler_angles = np.array([0.,np.pi/2,0.])
    rotated_coeff = soft.rotate_coeff_single(coeff,split_ids,euler_angles)
    
    expected = np.array([0.249745, 1.55, 0.285982, -1.45, 3.24974])
    
    is_close = np.isclose(rotated_coeff[4:9],expected,atol = 1e-5)
    if (~is_close).any():
        print("TEST FAILED!")
    else:
        print('test passed')
    return locals()

def test_calc_C():
    pi = np.pi
    bw=100
    shtns = shtns_plugin.sh(bw-1)
    soft = soft_plugin.Soft(bw)
    
    
    split_ids = shtns.cplx_l_split_indices
    ml_split_ids = shtns.cplx_m_split_indices
    lm_split_ids = shtns.cplx_l_split_indices
    
    sphere_grid = shtns.grid[:]
    half_sphere = np.zeros((128,)+sphere_grid.shape[:-1])
    #half_sphere[:]=1
    for i in range(99):
        half_sphere[:,i,i] = 1
        half_sphere[:,i,i+1] = 1
        half_sphere[:,i,i+2] = 1
        
    coeff = np.concatenate(shtns.forward_l(half_sphere),axis = 1)
    euler_angles = np.array([0.15,0.9625,1.9])*np.pi
    a,b,g = euler_angles/np.pi
    euler_angles2 = np.array([int(g)+1-(g-int(g)),b,int(a)+1-(a-int(a))])*np.pi
    coeff2 = soft.rotate_coeff(coeff,lm_split_ids,euler_angles)
    #coeff3 = soft.rotate_coeff_single(coeff2[0],lm_split_ids,euler_angles2)[None,:]
    #coeff31 = soft.rotate_coeff_single(coeff2[0],lm_split_ids,euler_angles3)[None,:]
    #coeff2 = np.concatenate(shtns.forward_l(i_half_sphere))[None,:]
    #coeff = np.zeros(shtns.n_coeff,np.complex128)
    #coeff[4:9]=1
    #coeff[4:9]=np.array([1.1, 0, 2, 3, 1])
    #coeff = np.random.rand(shtns.n_coeff) + 1j*np.random.rand(shtns.n_coeff)
    #coeff=coeff[None,:]
    
    #coeff2_lm = shtns.m_to_l_ordering(coeff2)
    #signal = np.random.rand(*shtns.grid.shape)
    #signal = signal[None,:]
    #coeff = shtns.forward_l(signal)
    #ignal = shtns.inverse_l(np.split(coeff,lm_split_ids,axis = 1))    

    
    #ml_coeff = np.concatenate([coeff[:,ids] for ids in shtns.cplx_m_indices],axis = 1)
    print('max_real coeff = {}'.format(coeff.real.max()))
    print('max_imax coeff = {}'.format(coeff.imag.max()))
    #C1 = soft.calc_mean_C(coeff2_lm[None,:],coeff2_lm[None,:],1,lm_split_ids)
    C2 = soft.calc_mean_C(coeff,coeff2,2,lm_split_ids)
    maxarg = np.unravel_index(np.argmax(C2),(2*bw,)*3)
    angles = soft.grid[maxarg[1],maxarg[0],maxarg[2]]/np.pi
    angles[0] = 2 - angles[0]
    angles[2] = 2 - angles[2]
    print('angles/pi = {}'.format(angles))
    return locals()



if __name__=='__main__':
    #locals().update(test_rotate_coeff())
    locals().update(test_calc_C())
    



import abc

class DetectorInterfaceSimulation(abc.ABC):
    @abc.abstractmethod
    def get_geometry(self):
        '''
        should return a dict containing the keys pixelGrid,pixelKornerIndex
        '''
        pass



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order,cross_correlation_low_pass
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward,circularHarmonicTransform_complex_inverse,circularHarmonicTransform_real_forward,circularHarmonicTransform_real_inverse
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import nearest_positive_semidefinite_matrix
from scipy import signal

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi,plot1D

def init_transforms():
    ccd= db.load('ccd')
    cc_data = ccd['cross_correlation'][...,:800]
    maxQ = np.max(ccd['radial_points'])
    max_order = 99
    n_radial_points = 128
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':0,'n_theta':0}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)    
    return locals()


def generate_density_spheres(grid,radius = 100, norm = 'standard',cart_centers= np.array([[0,200,0],[0,-200,0],[200,0,0],[-200,0,0],[400,0,0],[0,0,300]],dtype = np.float64),random_orientation = False):
    centers = mLib.cartesian_to_spherical(cart_centers)
    #centers = [(100,0,np.pi),(100,np.pi,np.pi),(100,0,0),(200,np.pi,0),(300,np.pi,np.pi/2)]
    sphere_funcs = [mLib.SampleShapeFunctions.get_disk_function(radius,center = center,norm=norm,random_orientation = random_orientation) for center in centers]
    density = sphere_funcs.pop(0)(grid)
    for sphere_func in sphere_funcs:
        density += sphere_func(grid)
    return density


def init():
    ccd = db.load('ccd')
    ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_model_0.h5')
    #sys._getframe(1).f_locals.update(init_transforms())
    #density = generate_density_spheres(grid_pair.real,radius = 200,norm='inf',cart_centers = np.array([[-200,0,0],[200,0,0]],dtype=np.float64),random_orientation = True)
    #density += generate_density_spheres(grid_pair.real,radius = 100,norm='inf',cart_centers = np.array([[0,300,0],[0,-100,0]],dtype=np.float64),random_orientation = False)
    #Bl = density_to_deg2_invariants_3d(density,cht,ft).real
    cc = ccd['cross_correlation']
    cc_1 = ccd_1['cross_correlation']
    grid = GridFactory.construct_grid('uniform',[ccd['qs'],ccd['qs']])

    max_order = 99
    proj_opt = settings.analysis.projections.reciprocal.pre_init.cross_correlation
    metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':1,'extract_odd_orders':False,'bl_enforce_psd':False}
    #cc_1,_locals = cross_correlation_low_pass(cc_1,max_order)
    qs = ccd['data_grid']['qs']
    #bl_single = cross_correlation_to_deg2_invariant(cc_1,3,**metadata)[0]
    #bl_single = nearest_positive_semidefinite_matrix(bl_single)

    low_pass_order = 120
    #sos = signal.butter(1, low_pass_order, 'lp', fs=128, output='sos')
    
    metadata_100 = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':100,'extract_odd_orders':False,'bl_enforce_psd':False}
    n_repeats = 1
    c_orders = np.concatenate((np.arange(cc.shape[-1]//2+1),-np.arange(1,cc.shape[-1]//2 + cc.shape[-1]%2)))
    c_odd_order_mask = c_orders%2
    new_cc = np.copy(cc)
    bls = []
    errors = []
    n_qs = cc.shape[0]
    n_phis = cc.shape[-1]
    periodic_cc=np.zeros((n_qs*2,n_qs*2,n_phis),dtype = cc.dtype)
    try:
        for n in range(n_repeats):
            cc_backup = np.copy(new_cc)
            cc_nonzero = cc_backup !=0
            if n < 0:
                periodic_cc[:n_qs,:n_qs]=new_cc
                periodic_cc[n_qs:,:n_qs]=new_cc[::-1,...]
                periodic_cc[:n_qs,n_qs:]=new_cc[:,::-1,:]
                periodic_cc[n_qs:,n_qs:]=new_cc[::-1,::-1,:]
                coeff_q1 = np.fft.rfft(periodic_cc,axis = 0)
                coeff_q1[-1]=0
                coeff_q1[-2]=0
                periodic_cc = np.fft.irfft(coeff_q1,n_qs*2,axis = 0)
                coeff_q2 = np.fft.rfft(periodic_cc,axis = 1)
                #coeff_q2[:,-2:]=0
                periodic_cc = np.fft.irfft(coeff_q2,n_qs*2,axis = 1)
                new_cc = periodic_cc[:n_qs,:n_qs,:n_phis]
                #new_cc = signal.sosfilt(sos,new_cc,axis = 0)
                #new_cc = signal.sosfilt(sos,new_cc,axis = 1)
            else:                
                pass
            cc_coeff = circularHarmonicTransform_real_forward(new_cc)        
            #cc_coeff[...,max_order+1:-max_order] = 0
            cc_coeff[...,max_order+1:] = 0
            #cc_coeff[...,c_odd_order_mask] = 0
            cc_coeff[...,1::2] = 0        
            new_cc = circularHarmonicTransform_real_inverse(cc_coeff,new_cc.shape[-1])
            cc_error = np.mean(np.abs((cc_backup - new_cc)[cc_nonzero])**2/np.abs(cc_backup[cc_nonzero])**2)
            print('cc_error = {}'.format(cc_error))

            if n == 0:
                bl = cross_correlation_to_deg2_invariant(new_cc,3,**metadata_100)[0]
            else:
                bl = cross_correlation_to_deg2_invariant(new_cc,3,**metadata)[0]                
            bl_backup = np.copy(bl)
            bl_nonzero = bl_backup !=0

            if n < -1:
                bl = signal.sosfilt(sos,bl,axis = -1)
                bl = signal.sosfilt(sos,bl,axis = -2)
            else:                
                pass
            bl = nearest_positive_semidefinite_matrix(bl)
            bl_error = np.mean(np.abs((bl_backup - bl)[bl_nonzero])**2/np.abs(bl_backup[bl_nonzero])**2)
            print('bl_error = {}'.format(bl_error))
            new_cc = deg2_invariant_to_cc_3d(bl,ccd['xray_wavelength'],ccd['data_grid'],np.arange(max_order+1))
            bls.append(bl)
            errors.append((cc_error,bl_error))
    except Exception as e :
        traceback.print_exc()
        print(e)
        bls = np.array(bls)
        errors = np.array(errors).T
    return locals()

locals().update(init())

def plot(orders):
    bl= np.abs(bls)
    error = errors
    datasets = [bl[0,orders],bl[0,orders],bl[0,orders]]

    n_orders = len(orders)
    
    layout_1 = {'title':'S','0 iteration':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_100 = {'title':'20 iteration','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout_A = {'title':'40 itertion','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    #layout_A2 = {'title':'M l-cc-phi psd l-bl-q','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    #layout_A3 = {'title':'M  l-cc-phi l-bl-q psd','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    #layout_A4 = {'title':'S psd','x_label':'$q_1$', 'y_label':'$q_2$','title_size':40,'text_size':30}
    layout = tuple(zip(*[[layout_1]*(n_orders+1),
                         [layout_100]*(n_orders+1),
                         [layout_A]*(n_orders+1),
                         #[layout_A2]*(n_orders+1),
                         #[layout_A3]*(n_orders+1),
                         #[layout_A4]*(n_orders+1),
                         ]))

    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/bl_loop_comparison.matplotlib',heat2D_multi.get_fig(tuple(zip(*datasets)),layout=layout, grid = grid,scale = 'log',shape = [n_orders,len(datasets)],size = (10*(len(datasets)+1),10*(n_orders)),vmin= 1e14, vmax = 1e24,cmap = 'plasma'))
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/bl_loop_errors.matplotlib',plot1D.get_fig(np.abs(error),y_scale = 'log'))

ccd = db.load('ccd')
cc = ccd['cross_correlation']
grid = GridFactory.construct_grid('uniform',[ccd['qs'],ccd['qs']])

db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/bl_loop_comparison_2.h5',{'bl':np.array(bls),'errors':np.array(errors).T})
data = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/bl_loop_comparison_2.h5')
bls = data['bl']
errors = data['errors']
orders = [2,4,6,8,10]    
plot(orders)





import time
import logging
import sys
import os
import math
import traceback
import numpy as np
import scipy as sp
import struct
from itertools import repeat
file_path = os.path.realpath(__file__)
plugin_dir = os.path.dirname(file_path)
os.chdir(plugin_dir)

from xframe.interfaces import ProjectWorkerInterface
from xframe import settings
from xframe import database
from xframe.library.pythonLibrary import xprint
from xframe.library.physicsLibrary import scattering_angle_to_reciprocal_radii
from xframe import Multiprocessing
from .projectLibrary.cross_correlation import ccf_analysis
Pi = np.pi
log=logging.getLogger('root')

def profile_modifier(func):
    def outer(*args,**kwargs):
        profile = settings.project.profile.use
        process_id=settings.project.profile.process_id
        if profile:
            if Multiprocessing.get_process_name() == process_id:
                out_folder=database.project.get_path('out_base',is_file=False)
                #log.info(f'out_folder = {out_folder}')
                database.project.create_path_if_nonexistent(out_folder)
                path = out_folder+"correlation_worker_{}-2.stats".format(process_id)
                import cProfile
                profiler = cProfile.Profile()
                profiler.enable()
        result = func(*args,**kwargs)
        if profile:
            if Multiprocessing.get_process_name() == process_id:
                profiler.disable()
                profiler.dump_stats(path)
        return result
    return outer
    

class ProjectWorker(ProjectWorkerInterface):
    def __init__(self):
        args = []
        arg_names = [
            'compute',
            'max_n_patterns',
            'batch_size',
            'n_processes',
            'list_inp',
            'image_dimensions',
            'intensity_pixel_threshold', 
            'intensity_radial_pixel_filter',
            'use_binary_mask',
            'background_subtraction',
            'ROI_normalization',
            'ROI_mean_filter',
            'pixel_size',
            'sample_distance',
            'wavelength',
            'polarization_correction',
            'solid_angle_correction',
            'qrange',
            'qrange_xcca',
            'fc_n_max',
            'phi_range',
            'ccf_2p_symmetrize',
            'detector_origin',
            'interpolation_order' 
        ]
        opt = settings.project
        if isinstance(opt.qrange,bool) or isinstance(opt.qrange_xcca,bool):
            wavelength = opt.wavelength
            sample_distance = opt.sample_distance
            detector_edge = opt.image_dimensions[0]/2*(opt.pixel_size/1000)
            scattering_angle = np.arctan(detector_edge/sample_distance)
            q_max = scattering_angle_to_reciprocal_radii(scattering_angle,wavelength)
            #qrange = (0,q_max,q_max/(opt.image_dimensions[0]/2-1))
            #qrange_xcca = ((0,q_max,1),(0,q_max,1))
            qrange = [0,q_max,q_max/(opt.image_dimensions[0]/2-1)]
            qrange_xcca = [[0,q_max,1],[0,q_max,1]]
            if isinstance(opt.qrange,bool):
                opt.qrange = qrange
            if isinstance(opt.qrange_xcca,bool):
                opt.qrange_xcca = qrange_xcca
            
        for n in arg_names:
            args.append(settings.project.get(n))

        
        args[4] = database.project.get_path('input_file_list')
        #log.info(f'len arg names - {len(arg_names)} len args = {len(args)}')            
        self.data_reader = DataReader(*args)
        
    def run(self):
        result = self.data_reader.run_processing_in_parallel()
        database.project.save('ccd',result)
        return {},locals()
    
# reading, processing and storing the data                
class DataReader():
    def __init__(self, compute, patterns_max, batch_size, numcpus_max,list_inp, image_dimensions,intensity_pixel_threshold, intensity_radial_pixel_filter, mask_binary, background_subtraction, ROInormalization, ROImeanfilter,pixsz, det_sam, wavelng, xpolarization, solid_angle_correction, qrange, qrange_xcca, fc_n_max, phirange, ccf_2p_symmetrize, dpcenter, interp_order):
        self.start_time = time.time()
        self.compute=compute
        self.patterns_max=patterns_max
        self.batch_size=batch_size
        self.numcpus_max=numcpus_max
        self.split_mode = settings.project.split_mode
        self.list_inp=list_inp
        self.intensity_pixel_threshold=intensity_pixel_threshold
        self.intensity_radial_pixel_filter=intensity_radial_pixel_filter
        self.mask_binary_inp=mask_binary
        self.background_subtraction=background_subtraction
        self.ROInormalization=ROInormalization
        self.ROImeanfilter=ROImeanfilter
        self.interp_order=interp_order
        self.fc_n_max=fc_n_max
        self.ccf_2p_symmetrize=ccf_2p_symmetrize
        self.pixelsize=pixsz
        self.det_sam=det_sam
        self.wavelng=wavelng
        self.dpcenter=dpcenter
        self.xpolarization=xpolarization
        self.solid_angle_correction=solid_angle_correction
        self.img_shape=image_dimensions
        
        # analyze dependencies of computations and update the "compute" list
        self._analyse_dependencies()
        
        #if os.path.exists(self.dir_save) is not True:
        #        os.makedirs(self.dir_save)
        
        #print('\nResults output directory: {}'.format(self.dir_save))
        print('Input file with a file list: {}'.format(self.list_inp))
         
        self.file_list=self._read_line_text(self.list_inp)
        self.M=len(self.file_list)
        print('Number of files specified in the list : {}'.format(self.M))
        cnt_r=0
        for i in range(self.M):
            if os.path.exists(self.file_list[i]):
                cnt_r+=1
            else:
                print('File {} does not exist '.format(self.file_list[i]))
        
        if cnt_r != self.M:
            print('Error: only {} files of {} specified in the list {} have been found.\nCorrect the input file list. \n'. format(cnt_r, self.M, self.list_inp))        
            sys.exit(1)
        
        self.good_frames=np.ones(self.M, dtype=int) # array with labels for each frame "1"-good, "0"-bad; initially we assume that all frames are good
        
        # read binary mask that will be applied to all images
        if self.mask_binary_inp==True:
            path = database.project.get_path('binary_mask')
            if os.path.exists(path):
                self.mask_binary=self._read_binary_2D_arr(path, self.img_shape)
            else:
                print('Error: Input file {} with binary mask have not been found.\n'.format(path))        
                sys.exit(1)
                
        # read background data from the file
        if self.background_subtraction:
            path = database.project.get_path('background')
            if os.path.exists(path):
                self.background_data=self._read_binary_2D_arr(path, self.img_shape)
            else:
                print('Error: Input file {} with background data have not been found.\n'. format(path))        
                sys.exit(1)
            
        # determime the reciprocal space geometry and related exp qunatities
        self._prepare_polar_representation(qrange, qrange_xcca, phirange) 
        
        
        # determine polarization and solid angle correction factors
        if self.xpolarization[0]:
            self._determine_polarization_correction()
        if self.solid_angle_correction==True:
            self._determine_solid_angle_correction()
                
        # define ROI for normalizing data
        if self.ROInormalization[0] or self.ROImeanfilter[0]:
            self.ROInorm_qpos1=np.abs(self.qvals - self.ROInormalization[1]).argmin()
            self.ROInorm_qpos2=np.abs(self.qvals - self.ROInormalization[2]).argmin()             
            if (self.ROInorm_qpos1==self.ROInorm_qpos2):
                print("WARNING: ROI normalization range ({},{}) contains only 1 radial point".format(self.qvals[self.ROInorm_qpos1],self.qvals[self.ROInorm_qpos2]))
            else:
                print("ROI normalization range ({},{}) contains {} radial points".format(self.qvals[self.ROInorm_qpos1],self.qvals[self.ROInorm_qpos2], self.ROInorm_qpos2-self.ROInorm_qpos1+1))    
        
        #initialize xcca functionality
        if "xcca" in self.compute:
            self.xcca_data=ccf_analysis(self.n_q1, self.n_q2, self.n_phi, self.q1vals_pos, self.q2vals_pos)
    
    
    # initiating multiprocessing, collecting the results
    #
    def run_processing_in_parallel(self):
        self.M=min(self.M, self.patterns_max)
        numcpus= Multiprocessing.get_free_cpus()
        self.numcpus=min(self.numcpus_max, numcpus)
    
        # create a list of batch indices
        #if 
        #batches = [(i, min(i + self.batch_size, self.M)) for i in range(0, self.M, self.batch_size)]
        #self.num_batches=len(batches)
        
        xprint("Number of CPUs available: {}, requested: {},  to be used: {} ".format(numcpus, self.numcpus_max, Multiprocessing._read_number_of_processes(self.numcpus)))
        xprint("Number of images to be processed: {}".format(self.M))
         
        # Process data batches in parallel
        mode = Multiprocessing.MPMode_Queue(assemble_outputs=False)
        results = Multiprocessing.process_mp_request(self.process_batch,mode,input_arrays=[np.arange(self.M)],n_processes = self.numcpus,split_mode=self.split_mode,call_with_multiple_arguments=True)
        
        if "is_good" in self.compute: 
                isgood_vals = np.zeros(self.M, dtype=int)
        if "waxs" in self.compute:    
                waxs_vals = np.zeros((self.M, self.n_q))
        if "waxs_aver" in self.compute:     
                waxs_aver_vals = np.zeros(self.n_q)
        if "xcca" in self.compute:   
            if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute):
                ccf_q1q2 = np.zeros((self.n_q1, self.n_q2, self.n_phi))
                ccf_q1q2_mask = np.zeros((self.n_q1, self.n_q2, self.n_phi), dtype=int)
        n_processes = len(results)
        log.info(f'len results = {len(results)}')
        split_ids = Multiprocessing.split_mp_arguments([np.arange(self.M)],n_processes,mode=self.split_mode)['indices']
        for i, batch in results.items():
            if not isinstance(batch,dict):
                batch = batch[0]
            #start_idx = i * self.batch_size
            #end_idx = min(start_idx + self.batch_size, self.M)
            #log.info(f'batch = {batch}')            
            if "is_good" in self.compute: 
                isgood_vals[split_ids[i]] = batch["is_good"]
            
            if "waxs" in self.compute:    
                waxs_vals[split_ids[i],:] = batch["waxs"]
            
            if "xcca" in self.compute: 
                if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute):   
                    np.add(ccf_q1q2, batch["ccf_q1q2"], out=ccf_q1q2)
                    np.add(ccf_q1q2_mask, batch["ccf_q1q2_mask"], out=ccf_q1q2_mask)
                
            
        Mgood=int(np.sum(isgood_vals))     #normalize properly        
        print("CCfs were averaged over {} selected images".format(Mgood))
        
        if "waxs_aver" in self.compute:    
            np.mean(waxs_vals, axis=0, where=(isgood_vals[:,None]==1), out=waxs_aver_vals)
                                
                
        if "xcca" in self.compute:      
            if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute):
                np.divide(ccf_q1q2, ccf_q1q2_mask, out=ccf_q1q2, where=(ccf_q1q2_mask!=0))
                ccf_q1q2[ccf_q1q2_mask==0]=np.nan
        
        if self.ccf_2p_symmetrize==True:
            posPi2=np.abs(self.phi - Pi/2.0).argmin()
            posPi=np.abs(self.phi - Pi).argmin()
            pos3Pi2=np.abs(self.phi - 3*Pi/2.0).argmin()
            ccf_q1q2[...]=self.xcca_data.symmetrize_ccf(ccf_q1q2, posPi2, posPi, pos3Pi2) 
            print("Two-point CCF has been symmetrized using the angular range (Pi/2, 3Pi/2)")
            
                   
        if "ccf_q1q2_fc" in self.compute:    
            ccf_q1q2_fc=self.xcca_data.ccf_fcs(ccf_q1q2)[...,:self.fc_n_max]
        

        # save results to output file
        result = {}
         
        if "waxs_aver" in self.compute:    
            prefix_save="iaverage"
            result['average_intensity'] = waxs_aver_vals
                  
        if "ccf_q1q2" in self.compute:
            prefix_save="ccf_q1q2_2p"
            result['cross_correlation'] = {'I1I1':ccf_q1q2}
            log.info(f'nans in cc = {np.isnan(ccf_q1q2).any()}')
            log.info(f'all nans = {np.isnan(ccf_q1q2).all()}')
            log.info(ccf_q1q2)
            
        if "ccf_q1q2_fc" in self.compute: 
            result['cross_correlation'] = {'I1I1_fc':ccf_q1q2_fc}
                        
        result['radial_points']=self.qvals
        result['angular_points']= self.phi#*(180.0/np.pi)
        result['num_images_processed']= self.M
        result['num_images_good'] = Mgood
        result['xray_wavelength']= settings.project.wavelength
        return result 
         

    
    # processing a batch of images in parallel
    #
    @profile_modifier
    def process_batch(self, frames,**kwargs):
        #read a batch of images from the files
        n_frames = len(frames)
        isgood_vals = np.zeros(n_frames)
        waxs_vals = np.zeros((n_frames,self.n_q),dtype = float)
        if "xcca" in self.compute:    
            if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute): 
                ccf_q1q2=np.zeros((self.n_q1, self.n_q2, self.n_phi))
                ccf_q1q2_mask = np.zeros((self.n_q1, self.n_q2, self.n_phi), dtype=int)
        if isinstance(self.batch_size,bool) or (self.batch_size ==0):
            n_parts = 1
        elif n_frames%self.batch_size!=0:
            n_parts = n_frames//self.batch_size+1
        else:
            n_parts = n_frames//self.batch_size
        n_parts = max(n_parts,1)
        
        batches = np.array_split(frames,n_parts)
        batch_ids = np.array_split(np.arange(len(frames)),n_parts)
        batch_nr = 1
        for batch,ids in zip(batches,batch_ids):
            data=np.empty((len(batch),self.img_shape[0], self.img_shape[1]))
            mask=np.ones(data.shape, dtype=int)
            for i,idx in enumerate(batch):
                #log.info(f'idx = {idx}')
                data[i,:,:]=self._read_binary_2D_arr(self.file_list[idx], self.img_shape)
            # perform processing of each frame in the batch
            for i,idx in enumerate(batch):            
                if self.good_frames[idx]==0: # if image was already labelled as bad
                    vals_dictionary = dict(zip(self.compute, repeat(0)))                    
                    if "is_good" in self.compute: 
                        vals_dictionary["is_good"]=0
                    if "xcca" in self.compute:    
                        vals_dictionary["image_polar"]=0
                        vals_dictionary["mask_polar"]=0    
                       
                else:
                    vals_dictionary=self.process_image(data[i], mask[i])
                    
                if "is_good" in self.compute: 
                    isgood_vals[ids[i]]=vals_dictionary["is_good"]
            
                if "waxs" in self.compute:    
                    waxs_vals[ids[i],:]=vals_dictionary["waxs"]
                        
                if (isgood_vals[i]==1) and  ("xcca" in self.compute):    
                    image_polar=vals_dictionary["image_polar"]
                    mask_polar=vals_dictionary["mask_polar"]
                    
                    if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute): 
                        valtmp,valmask=self.xcca_data.ccf_twopoint_q1_q2_mask_corrected(image_polar, mask_polar)
                        #isnotnone=~np.isnan(valtmp)
                        np.add(ccf_q1q2, valtmp.real, out=ccf_q1q2, where=valmask) 
                        np.add(ccf_q1q2_mask, 1, out=ccf_q1q2_mask, where=valmask)
            xprint(f'Process {kwargs["local_name"]}: Completed batch {batch_nr}/{n_parts}.')
            batch_nr+=1
        #return a combined result
        result_batch={}
        if "is_good" in self.compute: 
            result_batch["is_good"]=isgood_vals
            
        if "waxs" in self.compute:    
            result_batch["waxs"]=waxs_vals
       
        if "xcca" in self.compute: 
            if ("ccf_q1q2" in self.compute) or ("ccf_q1q2_fc" in self.compute):    
                result_batch["ccf_q1q2"]=ccf_q1q2
                result_batch["ccf_q1q2_mask"]=ccf_q1q2_mask
        
        #print("batch {} has been processed".format(frames))      
          
        return result_batch
    
    # processing an individual image 
    #
    def process_image(self, image, mask):
        
        vals_dictionary = dict(zip(self.compute, repeat(0))) # output results
        
        # update the mask if necessary
        if self.intensity_pixel_threshold[0]:    # mask pixels with intensities outside of the specified thresholds
                mask[(image<self.intensity_pixel_threshold[1]) | (image>self.intensity_pixel_threshold[2])]=0            
        if self.mask_binary_inp==True:
            np.multiply(mask, self.mask_binary, out=mask)
        
        # background correction
        if self.background_subtraction:
            image=np.subtract(image, self.background_data)
            
        #apply mask to the image 
        np.multiply(image, mask, out=image)
        
        # interpolate to polar coordinates
        image=sp.ndimage.map_coordinates(image, [self.cart_x.ravel(), self.cart_y.ravel()], order=self.interp_order, mode='constant', cval=0, prefilter=True)
        image.shape=(self.n_q, self.n_phi)
        mask=sp.ndimage.map_coordinates(mask, [self.cart_x.ravel(), self.cart_y.ravel()], order=self.interp_order, mode='constant', cval=0, prefilter=True)
        mask.shape=(self.n_q, self.n_phi)
        
        
        # radial pixel filter (can be applied only to the data in polar coordinates)
        if self.intensity_radial_pixel_filter[0]:
            if self.intensity_radial_pixel_filter[1][0]=='average_sigma':
                val_av, val_sig = self.i_average_and_sigma_azimuthal(image, mask)  
            elif self.intensity_radial_pixel_filter[1][0]=='median_mad':
                val_av, val_sig = self.i_median_and_mad(image, mask)
            else:
                print("Error: incorrect settings for intensity radial pixel filter. The filter will not function properly")
            
            
            #apply filter
            mask[np.abs(image-val_av[:,None])>self.intensity_radial_pixel_filter[1][1]*val_sig[:,None]]=0
            np.multiply(image, mask, out=image)      
            
                               
        is_good=1 # initially consider all images as good
        
        if (np.sum(mask)==0):
            print("found a completely masked image")
            vals_dictionary["is_good"]=0 #update dictionary
            return vals_dictionary
        
                 
        if self.ROImeanfilter[0] or self.ROInormalization[0] or ("mean_roi" in self.compute):
            meav_roi_val=np.mean(image[self.ROInorm_qpos1:self.ROInorm_qpos2], where=(mask[self.ROInorm_qpos1:self.ROInorm_qpos2]==1)) # average intensity/pixel
        
        if self.ROImeanfilter[0]:
            if (meav_roi_val<self.ROImeanfilter[1]) or (meav_roi_val>self.ROImeanfilter[2]):
                is_good=0
                
        if self.ROInormalization[0]: 
            image=np.divide(image, meav_roi_val)
                
        if self.xpolarization[0]==True:
                image=np.multiply(image, self.Pfactor)
                
        if self.solid_angle_correction==True:
                image=np.multiply(image, self.SolAngCorr)

                
        #compose results
        if "is_good" in self.compute: 
            vals_dictionary["is_good"]=is_good
        
        if "waxs" in self.compute: 
            val_av = self.i_average_azimuthal(image, mask)   
            vals_dictionary["waxs"]=val_av
        if "xcca" in self.compute:    
            vals_dictionary["image_polar"]=image
            vals_dictionary["mask_polar"]=mask
                  
        return  vals_dictionary 
    
   
    
    # compute the mean value and standard deviation for an image determined on a polar grid
    #
    def i_average_and_sigma_azimuthal(self, data_polar, mask_polar):
        imean=np.mean(data_polar, axis=1, where=(mask_polar==1))
        istd=np.std(data_polar, axis=1, where=(mask_polar==1))
        return imean, istd 
    
    # compute the mean value for an image determined on a polar grid
    #
    def i_average_azimuthal(self, data_polar, mask_polar):
        imean=np.mean(data_polar, axis=1, where=(mask_polar==1))
        return imean
        
    # compute median value and median absolute deviation for an image determined on a polar grid
    #
    def i_median_and_mad(self, data_polar, mask_polar):
        imedval=np.nanmedian(np.where(mask_polar==1, data_polar, np.nan),axis=1)
        imad=spst.median_abs_deviation(np.where(mask_polar==1, data_polar, np.nan), axis=1, nan_policy='omit')
        return imedval, imad    
                    
    # analyze dependencies and update the compute list
    #
    def _analyse_dependencies(self):
        if ("ccf_q1q2" in self.compute): 
            if ("xcca" not in self.compute): 
                self.compute.append("xcca")         
        if "waxs_aver" in self.compute:
            if not ("waxs" in self.compute):
                 self.compute.append("waxs")
   
                
    # reciprocal space geometry
    #
    def _prepare_polar_representation(self, qrange, qrange_xcca, phirange):
    
        self.q_min=qrange[0]
        self.q_max=qrange[1]
        self.q_step=qrange[2]
        pixsz=self.pixelsize*0.001 # conversion to [mm]
        self.n_q=int((self.q_max-self.q_min)/self.q_step+1)    # number of points in the raial direction (radial sampling)
        self.phi_min=phirange[0]
        self.phi_max=phirange[1]
        n_phi=phirange[2]
        
        detq=(2*Pi*pixsz)/(self.wavelng*self.det_sam) #  Detector resolution in reciprocal space [ Angstroem^(-1)] 
       
        self.qvals=np.arange(self.n_q)*self.q_step + self.q_min    # qvals for each ring
        self.theta=2.0*np.arcsin(self.qvals*self.wavelng/(4.0*Pi)) # scattering angle 2theta for each q-ring
      
        # maximum feasible azimuthal sampling
        min_circ_rad_pix=int(math.floor(self.det_sam*math.tan(self.theta[0])/pixsz)) # circle radius [pixels] determined for q_min
        max_circ_rad_pix=int(round(self.det_sam*math.tan(self.theta[-1])/pixsz)) # circle radius [pixels] determined for q_max
        step_circ_rad_pix=self.det_sam*math.tan(self.theta[1]-self.theta[0])/pixsz # circle radius [pixels] determined for q_step (at the lowest resoltion)
        
        self.maxpix=round(2*Pi*max_circ_rad_pix) # number of pixels on the circumference of the radius q_max, determines the maximum feasible azimuthal sampling (1 pixel step)
        if (self.maxpix % 2) != 0:
            self.maxpix+=1
        self.maxpix=int(self.maxpix)
        
        
        # number of points in the azimuthal direction(azimuthal sampling) to be used, the same for all resolution rings
        if phirange[3]=='max':
            self.n_phi=min(self.maxpix, n_phi)
        elif phirange[3]=='min':
            self.n_phi=max(self.maxpix, n_phi)
        else:
            self.n_phi=n_phi
                      
        self.phi_step=(self.phi_max-self.phi_min)/self.n_phi
        self.phi=np.arange(self.n_phi)*(self.phi_max-self.phi_min)/float(self.n_phi) + self.phi_min # azimuthal angles for each point
             
        print("Reciprocal space parameters: q_min={:.6f} [A^-1], q_max={:.6f} [A^-1], q_step={:.6f} [A^-1], N_q={}".format(self.qvals[0], self.qvals[-1], self.q_step, self.n_q))
        print("Reciprocal space parameters: q_min={} [pixels], q_max={} [pixels], q_step={:.3f} [pixels]".format(min_circ_rad_pix, max_circ_rad_pix, step_circ_rad_pix))
        
        if (self.q_step<detq):
            print("WARNING: the specified q_step={:.6f} is smaller than the detector resolution detq={:.6f} (q_step-detq={:.8f}). To save computation time set q_step>=detq".format(self.q_step, detq, self.q_step-detq))
        print("Scattering angles: 2theta_min={:.4f} [deg], 2theta_max={:.4f} [deg], 2theta_step={:.4f} [deg]".format(self.theta[0]*180.0/Pi, self.theta[-1]*180.0/Pi, (self.theta[1]-self.theta[0])*180.0/Pi))
        print("Maximum feasible number of azimuthal points at q_max: max_n_phi={}; requested : {}; will be used={}".format(self.maxpix, n_phi, self.n_phi)) 
  
        #polar grid [pixel coordinates]
        self.pol_q = np.outer(np.tan(self.theta)*self.det_sam/pixsz, np.ones(self.n_phi) )
        self.pol_phi = np.outer( np.ones(self.n_q), np.arange(self.n_phi)*(self.phi_max-self.phi_min)/float(self.n_phi) + self.phi_min)
        
        # cartesian coordinates
        self.cart_x_orig = self.pol_q*np.cos(self.pol_phi)
        self.cart_y_orig = self.pol_q*np.sin(self.pol_phi)
        
        self.cart_x = self.cart_x_orig + self.dpcenter[0]
        self.cart_y = self.cart_y_orig + self.dpcenter[1]
            
        # computations of the [q1,q2] ranges for xcca
        q1pos1=np.abs(self.qvals - qrange_xcca[0][0]).argmin()# find the positons of the q1_min and q1_max in the array  self.qvals
        q1pos2=np.abs(self.qvals - qrange_xcca[0][1]).argmin()
        self.q1vals_pos=np.arange(q1pos1,q1pos2+1,qrange_xcca[0][2]) # indices of the q1 values in the array of self.qvals
        self.q1vals=self.qvals[self.q1vals_pos]
        self.n_q1=self.q1vals.shape[0]
        print("q1 values for computing the CCFs: q1_min={:.6f} [A^-1], q1_max={:.6f} [A^-1], q1_step={:.6f} [A^-1], N_q1={}".format(self.q1vals[0], self.q1vals[-1], self.q1vals[1]-self.q1vals[0], self.n_q1))
        
        q2pos1=np.abs(self.qvals - qrange_xcca[1][0]).argmin()# find the positons of the q1_min and q1_max in the array  self.qvals
        q2pos2=np.abs(self.qvals - qrange_xcca[1][1]).argmin()
        self.q2vals_pos=np.arange(q2pos1,q2pos2+1,qrange_xcca[1][2]) # indices of the q1 values in the array of self.qvals
        self.q2vals=self.qvals[self.q2vals_pos]
        self.n_q2=self.q2vals.shape[0]
        print("q2 values for computing the CCFs: q2_min={:.6f} [A^-1], q2_max={:.6f} [A^-1], q2_step={:.6f} [A^-1], N_q2={}".format(self.q2vals[0], self.q2vals[-1], self.q2vals[1]-self.q2vals[0], self.n_q2))
        
  
    # determine polarization factor
    # the experimental intensity should be multiplied by self.Pfactor[i,j]
    #
    def _determine_polarization_correction(self):
        
        self.Pfactor=np.ones((self.n_q, self.n_phi))
        
        # here 'h' and 'v' polarisations are implemented in the way to be compatible with array indexing (firt index - verical direction, second, horisontal)
        if self.xpolarization[1]=='v':
          for i in range(self.n_q):
            th=self.theta[i]
            for j in range(self.n_phi):
                fi=self.phi[j]
                self.Pfactor[i,j]=1.0/(math.cos(th)**2+(math.sin(th)**2)*(math.sin(fi)**2))

        elif self.xpolarization[1]=='h':
          for i in range(self.n_q):
            th=self.theta[i]
            for j in range(self.n_phi):
                fi=self.phi[j]
                self.Pfactor[i,j]=1.0/(math.cos(th)**2+(math.sin(th)**2)*(math.cos(fi)**2))
        
    # determine solid_angle_correction factor
    # the experimental intensity should be multiplied by self.SolAngCorr[i] 
    #
    def _determine_solid_angle_correction(self):
        self.SolAngCorr=np.empty((self.n_q, self.n_phi)) # multiply raw intensity by this factor 
        for i in range(self.n_q):
            th=self.theta[i]
            self.SolAngCorr[i]=1.0/(math.cos(th)**3)
    
    # read text file content line-by-line
    def _read_line_text(self, fname):
        with open(fname) as f:
            text_lines = f.readlines()
            text_lines = [x.strip() for x in text_lines] 
            text_lines = list(filter(None, text_lines))
            base_path = os.path.dirname(database.project.get_path('input_file_list'))
            for _id,f in enumerate(text_lines):
                if f[:2]=='./':
                    text_lines[_id]=os.path.join(base_path,f[2:])
            return text_lines

    # read 2D binary array of a given shape from a file
    def _read_binary_2D_arr(self, fname, shape, dtype='f', bo='<' ): 
        with open( fname, "rb") as file:
            data = file.read()
        if len(data)!=0:  
            fmt=bo+str(shape[0]*shape[1])+dtype
            dtuple=struct.unpack(fmt, data)
            data=np.asarray(dtuple).reshape(shape)
            #data = database.project.load(fname).reshape(shape)
            log.info(f'data contains {np.isnan(data).sum()} nan values')
            data[np.isnan(data)]=0
            #log.info(f'datatype = {data.dtype}')
            #data = data.astype(float)
            #log.info(f'data = {data[0][:3]}')
            #log.info(f'data db = {database.project.load(fname)[:3]}')
            return data
        else:
            return np.array([]) 
            



from xframe._version import __version__

#make sure  numpy does not try to multiprocess over xframes multiprocessing
import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"

from xframe import startup_routines

startup_routines.import_settings()
startup_routines.setup_logging()
import logging
log = logging.getLogger('root')
startup_routines.startup_imports()
_external_dependency_lookup = startup_routines.create_dependency_dict()

startup_routines.dependency_injection()
dependency_inject_SOFT = startup_routines.dependency_inject_SOFT
startup_routines.setup_default_database()


# Initialize attributes #
from xframe.interfaces import ExperimentWorkerInterface,ProjectWorkerInterface
from xframe.control.Control import Controller
#The following variables will be set on runtime
_project_worker_module_name = False  # will be a string
_experiment_module_name = False      # will be a string
project_worker = ProjectWorkerInterface
experiment = ExperimentWorkerInterface
controller = Controller()

startup_routines.setup_projects()
startup_routines.setup_experiments()
select_project = startup_routines.select_project
select_experiment = startup_routines.select_experiment
import_selected_project = startup_routines.import_selected_project
import_selected_experiment= startup_routines.import_selected_experiment
run = startup_routines.run
select_and_run = startup_routines.select_and_run
reload_home = startup_routines.reload_home
change_home = startup_routines.change_home




import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward

import logging
#from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder2D,AgipdRegridderSimple

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

log=logging.getLogger('root')
db = default_DB()

init = False
if init:
    from xframe.startup_routines import dependency_injection
    dependency_injection()
    from xframe.startup_routines import load_recipes
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='mask_data_high',exp_name='fxs3046_online',exp_settings = 'default_proc')
    controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
    controller.chooseJobSingleProcess()
    from xframe import database
    from xframe import Multiprocessing



from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.gridLibrary import GridFactory
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library import units
from xframe.library.gridLibrary import get_linspace_log2
from xframe.library.gridLibrary import uniformGrid_func

from scipy.spatial.kdtree import KDTree
from scipy.spatial import Delaunay 
from scipy.ndimage import gaussian_filter
from scipy.ndimage import spline_filter

def create_lookup_array_2D(grid,new_grid,return_counts = False,coord_sys='cartesian'):
    '''assumes grids are in cartesian_coordinates'''
    if coord_sys == 'spherical':        
        grid = spherical_to_cartesian(grid[:])
        new_grid = spherical_to_cartesian(new_grid[:])
    if grid.shape[-1] == 3:
        grid = grid[...,:2]
    shape = grid.shape
    new_shape = new_grid.shape
    tree = KDTree(grid.reshape(-1,2))
    lookup_ids = tree.query(new_grid.reshape(-1,2))[1]
    unraveld_lookup = np.unravel_index(lookup_ids,shape[:-1])
    unraveld_lookup = tuple(part.reshape(new_shape[:-1]) for part in unraveld_lookup)
        
    if return_counts:
        counts = np.zeros(new_shape[0])
        shaped_lookup = lookup_ids.reshape(new_shape[:-1])
        for n,x in enumerate(shaped_lookup):
            counts[n] = len(np.unique(x))
        return unraveld_lookup,counts
    else:
        return unraveld_lookup
    

class SimpleRegridder2D:
    def __init__(self,pixel_centers,new_grid,interpolation = 'nearest',coord_sys='spherical',interpolation_constants=False):
        ''' assumes pixel_centers in spherical coordinates'''
        self.out_shape = new_grid.shape[:-1]
        self.in_shape = pixel_centers.shape[:-1]
        self.pixel_centers = pixel_centers
        self.new_grid = new_grid
        self.interpolation = interpolation
                
        if coord_sys == 'spherical':
            self.pixel_centers = spherical_to_cartesian(pixel_centers)
            self.new_grid = spherical_to_cartesian(new_grid)
        self.pixel_centers_hash = hash(pixel_centers.data.tobytes())
        self.new_grid_hash = hash(new_grid.data.tobytes())
        
        try:
            self.interpolation_constants = getattr(self,'generate_interpolation_constants_'+interpolation)(interpolation_constants)
            self.apply = getattr(self,'generate_apply_'+interpolation)()
        except AttributeError as e:
            log.error('interpolation method {} not known. Available interpolation options are {}'.format(interpolation,'"nearest","linear"')) 

    def check_interpolation_constants(self,interpolation_constants):

        pixel_centers_hash = hash(self.pixel_centers.data.tobytes())
        new_grid_hash = hash(self.new_grid.data.tobytes())
        pixel_centers_match = pixel_centers_hash == interpolation_constants['hash_pixel_centers']
        new_grids_match = new_grid_hash == interpolation_constants['hash_new_grid']
        interpolation_match = self.interpolation == interpolation_constants['interpolation']
        constants_valid = (new_grids_match and pixel_centers_match and interpolation_match)
        #print('constants valid',constants_valid)
        #print('new_grid_match',new_grids_match)
        return constants_valid
    
    def assemble_interpolation_dict(self, data: dict):
        interpolation_constants={
            'hash_pixel_centers':hash(self.pixel_centers.data.tobytes()),
            'hash_new_grid':hash(self.new_grid.data.tobytes()),
            'interpolation':self.interpolation,
            **data
        }
        return interpolation_constants

    def generate_interpolation_constants_nearest(self,interpolation_constants):
        constants_fit_to_grids = False
        if isinstance(interpolation_constants,dict):
            constants_fit_to_grids = self.check_interpolation_constants(interpolation_constants)
        if not constants_fit_to_grids:
            lookup_array,counts = create_lookup_array_2D(self.pixel_centers,self.new_grid,return_counts = True)
            data = {'lookup_array':lookup_array}
            interpolation_constants = self.assemble_interpolation_dict(data)
        return interpolation_constants    
    def generate_interpolation_constants_linear(self,interpolation_constants):
        constants_fit_to_grids = False
        if isinstance(interpolation_constants,dict):
            constants_fit_to_grids = self.check_interpolation_constants(interpolation_constants)
        if not constants_fit_to_grids:
            vertices,weights = self.generate_linear_interp_weights(self.pixel_centers,self.new_grid)            
            data = {'vertices':vertices,'weights':weights}
            interpolation_constants = self.assemble_interpolation_dict(data)            
        return interpolation_constants

    def generate_linear_interp_weights(self,grid,new_grid):
        d = grid.shape[-1]
        g = grid.reshape(-1,d)
        ng = new_grid.reshape(-1,d)
        log.info('grid shape = {} new grid shape = {}'.format(g.shape,ng.shape))
        tri = Delaunay(g)
        simplex = tri.find_simplex(ng)
        outside_hull = (simplex == -1)
        vertices = tri.simplices[simplex]
        temp = tri.transform[simplex]
        delta = ng - temp[:, d]
        bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
        weights = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))
        weights[outside_hull] = 0
        return vertices, weights

    
    def generate_apply_linear(self):
        weights = self.interpolation_constants['weights']
        vertices = self.interpolation_constants['vertices']
        shape = self.out_shape
        ndim_in = len(self.in_shape)
        def apply(data):
            ndim = data.ndim
            N=data.shape[-1]
            if ndim == 1+ndim_in:
                return np.einsum('njk,nj->nk', data.reshape(-1,N)[vertices], weights).reshape(shape+(N,))
            else:                
                return np.einsum('nj,nj->n', data.reshape(-1)[vertices], weights).reshape(shape)
        return apply
    
    def generate_apply_nearest(self):
        lookup = self.interpolation_constants['lookup_array']
        def apply(data):
            if data.ndim == 1+len(lookup):
                return data[lookup]
            else:
                return data[lookup]
        return apply

class AgipdRegridderSimple:
    version = 0.1
    @staticmethod
    def get_interpolation_constants_name(geometry,new_grid_shape,interpolation):
        pixel_hash = hash(geometry['framed_pixel_centers'].data.tobytes())
        new_grid_hash = hash(new_grid_shape + (AgipdRegridderSimple.version,))
        return '{}_{}_{}'.format(interpolation,pixel_hash,new_grid_hash)

    
    def __init__(self,geometry,new_grid_shape, interpolation = 'nearest', interpolation_constants=False):        
        spherical_pixel_centers = geometry['framed_pixel_centers']
        self.default_interpolation_constants = {
            'data':{m_id:False for m_id in range(len(spherical_pixel_centers))},
            'mask':{m_id:False for m_id in range(len(spherical_pixel_centers))}
        }
        self.geometry = geometry
        self.sensitve_pixel_mask = geometry['framed_mask']
        self.interpolation = interpolation
        self.new_grid_shape = new_grid_shape
        self.max_pixel_q = np.max(spherical_pixel_centers[...,0])
        self.new_constants_generated = False
        self.generate_new_constants,self.interpolation_constants = self.process_interpolation_constants(interpolation_constants)
        polar_pixel_centers = spherical_pixel_centers[...,::2]
        cart_pixel_centers = spherical_to_cartesian(polar_pixel_centers)


        module_boundaries = []
        for dim in range(2):
            _min = np.min(cart_pixel_centers[:,...,dim],axis = (1,2))
            _max = np.max(cart_pixel_centers[:,...,dim],axis = (1,2))
            module_boundaries.append([_min,_max])        
        self.module_boundaries = np.moveaxis(np.array(module_boundaries),-1,0)
        
        self.new_grid = self._generate_new_grid(new_grid_shape,self.max_pixel_q)
        
        self.module_masks = self._generate_module_masks(self.new_grid,self.module_boundaries)

        temp = self._generate_regridders(self.new_grid,polar_pixel_centers,self.module_masks)
        self.data_regridders_per_module = temp[0]
        self.mask_regridders_per_module = temp[1]

        if self.generate_new_constants:
            self.interpolation_constants = self._assemble_interpolation_constants()
            self.new_constants_generated = True

        self.regrid = self.generate_regrid()

    def _generate_new_grid(self,new_grid_shape,max_q):
        Nq,Nphi = new_grid_shape
        qs = max_q*np.arange(Nq)/(Nq-1)
        phis = 2*np.pi*np.arange(Nphi)/Nphi
        new_grid = GridFactory.construct_grid('uniform',[qs,phis])[:]
        return new_grid

    def _generate_module_masks(self,new_grid,boundaries):
        cart_grid = spherical_to_cartesian(new_grid)

        min_mask = cart_grid>=boundaries[:,None,None,:,0]
        max_mask = cart_grid<=boundaries[:,None,None,:,1]
        self.min_mask = min_mask
        self.max_mask = max_mask
        module_masks = min_mask[...,0] & min_mask[...,1] & max_mask[...,0] & max_mask[...,1]        
        return module_masks
    
    def _generate_regridders(self,new_grid,polar_pixel_centers,module_masks):
        constants = self.interpolation_constants                
        data_regridders = []
        mask_regridders = []        
        for m_id,mask in enumerate(module_masks):
            module_pixel_centers = polar_pixel_centers[m_id]
            new_points = new_grid[mask]
                
            r = SimpleRegridder2D(module_pixel_centers,new_points,interpolation = self.interpolation,interpolation_constants = constants['data'][m_id])
            if self.interpolation == 'linear':
                rm = SimpleRegridder2D(module_pixel_centers,new_points,interpolation = 'nearest',interpolation_constants = constants['mask'][m_id])
            else:
                rm = r
            data_regridders.append(r)
            mask_regridders.append(rm)
        
        return data_regridders,mask_regridders
    def _assemble_interpolation_constants(self):
        data_constants = {m_id:r.interpolation_constants for m_id,r in enumerate(self.data_regridders_per_module)}
        mask_constants = {m_id:r.interpolation_constants for m_id,r in enumerate(self.mask_regridders_per_module)}
        hash_name = self.get_interpolation_constants_name(self.geometry,self.new_grid_shape,self.interpolation)        
        interpolation_constants={'data' : data_constants,'mask':mask_constants,'hash_name':hash_name}
        return interpolation_constants
    def process_interpolation_constants(self,constants):
        constants_are_valid = False
        if isinstance(constants,dict):            
            hash_name = self.get_interpolation_constants_name(self.geometry,self.new_grid_shape,self.interpolation)        
            constants_name = constants['hash_name']
            constants_are_valid = (hash_name == constants_name)
            
        if constants_are_valid:
            interpolation_constants = constants
        else:
            interpolation_constants = self.default_interpolation_constants
        return (not constants_are_valid),interpolation_constants
    def generate_regrid(self):
        in_mask = self.sensitve_pixel_mask
        in_shape = self.sensitve_pixel_mask.shape
        out_shape = self.new_grid_shape
        module_masks = self.module_masks
        data_regridders = self.data_regridders_per_module
        mask_regridders = self.mask_regridders_per_module
        def regrid(data,mask,modules,**args):
            ndim = data.ndim
            if ndim == len(in_mask):
                N = len(data)
                # have to copy data to framed agipd shape
                data_dtype = data.dtype
                mask_dtype = mask.dtype
                in_d = np.zeros(in_shape+(N,),data_dtype)
                in_m = np.zeros(in_shape+(N,),mask_dtype)
                out_d = np.zeros((N,)+out_shape,data_dtype)
                out_m = np.zeros((N,)+out_shape,mask_dtype)
                #print(in_d[:,in_mask].shape)
                #print(data.shape)            
                in_d[in_mask] = np.moveaxis(data,0,-1).reshape(-1,N)
                in_m[in_mask]= np.moveaxis(mask,0,-1).reshape(-1,N)
                for m in modules:
                    mask = module_masks[m]
                    out_d[:,mask]= np.moveaxis(data_regridders[m].apply(in_d[m]),-1,0)
                    out_m[:,mask]= np.moveaxis(mask_regridders[m].apply(in_m[m]),-1,0)                
            else:
                data_dtype = data.dtype
                mask_dtype = mask.dtype
                in_d = np.zeros(in_shape,data_dtype)
                in_m = np.zeros(in_shape,mask_dtype)
                out_d = np.zeros(out_shape,data_dtype)
                out_m = np.zeros(out_shape,mask_dtype)
                in_d[in_mask] = data.reshape(-1)
                in_m[in_mask]= mask.reshape(-1)
                for m in modules:
                    mask = module_masks[m]
                    out_d[mask]= data_regridders[m].apply(in_d[m])
                    out_m[mask]= mask_regridders[m].apply(in_m[m])                                
            return out_d,out_m
        return regrid
                
                


    
comm = Multiprocessing.comm_module
g = comm.get_geometry()

base_path = '/gpfs/exfel/theory_group/user/berberic/fxs3046/test/regrid2/'

pixel_centers = g['framed_pixel_centers'][:,::16,::16]
sensitive_pixel = g['framed_mask'][:,::16,::16]
data = np.zeros((2,16,32,8))
for module in data:
    for n,i in enumerate(module):
        i[:,n%2::2]=1.0
mask = np.full(data.shape,True,dtype = bool)

from scipy.interpolate import griddata

#cart_grid = GridFactory.construct_grid('uniform',[np.arange(-20,20,dtype = float)/5,np.arange(-20,20,dtype = float)/5])[:]
#polar_grid = GridFactory.construct_grid('uniform',[np.arange(40,dtype = float)/4,np.arange(1000)*2*np.pi/1000])[:]
#rg = SimpleRegridder2D(cart_grid,spherical_to_cartesian(polar_grid),interpolation='linear',coord_sys='cartesian')
#rg2 = SimpleRegridder2D(spherical_to_cartesian(polar_grid),cart_grid,interpolation='linear',coord_sys='cartesian')
#
#data = np.zeros((100,100))
#data[::2,1::2]=1
#
#data_polar = np.zeros(polar_grid.shape[:-1])
#data_polar[::2]=1
#new_data = rg2.apply(data_polar)
#polar_cart = spherical_to_cartesian(polar_grid)
#new_data2 = griddata((polar_cart[...,0].reshape(-1),polar_cart[...,1].reshape(-1)),data_polar.reshape(-1),(cart_grid[...,0].reshape(-1),cart_grid[...,1].reshape(-1)),fill_value = 0.0).reshape(new_data.shape)

#pixel_centers[...,1]+=np.pi/2
#pixel_centers[...,2]+=np.pi
gg = {'framed_pixel_centers':pixel_centers,'framed_mask':sensitive_pixel}
new_grid_shape = (512,1024*2)
ar = AgipdRegridderSimple(gg,new_grid_shape,interpolation='nearest',interpolation_constants=ic)


polar_data,polar_mask = ar.regrid(data[0],mask[0],np.arange(8))

#new_data = np.zeros_like(ar.new_grid[...,0])
#for mask in ar.module_masks:
#    new_data[mask]=1

base_path = '/gpfs/exfel/theory_group/user/berberic/p3046/tests/'
database.analysis.save(base_path+'regrid_simple.vts',[polar_data],grid=ar.new_grid,grid_type='polar')
#database.analysis.save(base_path+'regrid_simple3.vtr',[new_data2],grid=cart_grid,grid_type='cartesian')
#database.analysis.save(base_path+'regrid_simple2.vtr',[new_data],grid=cart_grid,grid_type='cartesian')





import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi,plot1D
from xframe import Multiprocessing
from scipy.interpolate import interp1d
from scipy.linalg import solve_triangular

log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
from xframe.library.mathLibrary import gsl
#analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_10p_spher')[0] 
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)


ccd = db.load('ccd')
ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_model_0.h5')
#ccd_1 = db.load_ccd('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_pentagon_10particles_rectSim.h5')
cc = ccd['cross_correlation']
cc_1 = ccd_1['cross_correlation']

pre_init_opt = settings.analysis.projections.reciprocal.pre_init
proj_opt = pre_init_opt.cross_correlation

def init_transforms():
    cc_data = ccd_1['cross_correlation'][...,:800]
    maxQ = np.max(ccd_1['radial_points'])*2
    max_order = 99
    n_radial_points = 128#300
    opt=settings.analysis
    ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
    #ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
    # harmonic transforms and grid
    cht=HarmonicTransform('complex',ht_opt)
    weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points,expansion_limit = 2000)
    grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':maxQ,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
    grid_pair = get_grid(grid_opt)
    maxR = grid_pair.real[:,0,0].max()
    ft,ift = generate_zernike_spherical_ft(maxR,weight_dict,cht,'complex',True,use_gpu=True)    
    return locals()

#locals().update(init_transforms())

def generate_density_spheres(grid,radius = 100, norm = 'standard',cart_centers= np.array([[0,200,0],[0,-200,0],[200,0,0],[-200,0,0],[400,0,0],[0,0,300]],dtype = np.float64),random_orientation = False):
    centers = mLib.cartesian_to_spherical(cart_centers)
    #centers = [(100,0,np.pi),(100,np.pi,np.pi),(100,0,0),(200,np.pi,0),(300,np.pi,np.pi/2)]
    sphere_funcs = [mLib.SampleShapeFunctions.get_disk_function(radius,center = center,norm=norm,random_orientation = random_orientation) for center in centers]
    density = sphere_funcs.pop(0)(grid)
    for sphere_func in sphere_funcs:
        density += sphere_func(grid)
    return density


def init2():
    r = 65
    phis = np.arange(5)*2*np.pi/5
    centers = 2*r*np.stack((np.sin(phis),np.cos(phis),np.zeros(5)),axis = -1)
    centers = list(centers)
    centers.append(np.array([0.0,0.0,0.0]))
    centers = np.array(centers)
    centers2 = np.array([centers[0],centers[2],centers[3]])
    density1 = generate_density_spheres(grid_pair.real,radius = r,cart_centers=centers)
    density2 = generate_density_spheres(grid_pair.real,radius = r,cart_centers=centers2)
    density = 25*(density1+density2)
    #bls_exact = density_to_deg2_invariants_3d(density,cht,ft)
    
    return locals()

#locals().update(init2())

def plot2():
    orders = np.arange(0,10,2)
    #orders = np.arange(10,100,10)
    #orders = [2]
    qs = grid_pair.reciprocal[:,0,0,0]
    qq_grid = GridFactory.construct_grid('uniform',[qs,qs])
    bls = bls_exact
    
    fig = heat2D_multi.get_fig(list(np.abs(bls[orders])),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,5),vmin = 1e+14, vmax = 1e+26,scale='log',cmap='plasma')
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact0-10.matplotlib',fig)

    fig = heat2D_multi.get_fig(list(np.abs(mLib.nearest_positive_semidefinite_matrix(bls[orders]))),grid = qq_grid,shape = (1,len(orders)),size=(5*len(orders)+2*5,5),vmin = 1e+14, vmax = 1e+26,scale='log',cmap='plasma')
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact_psd0-10.matplotlib',fig)

    intensity_exact = (np.abs(ft(density))**2).real
    Ilm = cht.forward(intensity_exact)
    fig = plot1D.get_fig(np.abs(Ilm[2][:,0]),y_scale='log',grid = qs)
    db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/Ilm_exact.matplotlib',fig)
    #bl2_new = mLib.nearest_positive_semidefinite_matrix(bls[...,2])
    #fig = heat2D.get_fig(np.abs(bl2_new.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2n.matplotlib',fig)
    #bl2_new = (bls[...,2]+bls[...,2].T)/2
    #fig = heat2D.get_fig(np.abs(bl2_new.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2nn.matplotlib',fig)
    #
    #s,v = np.linalg.eigh(bl2_new)
    #s[s<s.max()*1e-15]=0
    #bl2_new2 = v*s[None,:] @ v.T 
    #fig = heat2D.get_fig(np.abs(bl2_new2.real),scale='log')
    #db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/bl_exact2nnn.matplotlib',fig)

#plot2()



#rd = ft(density)
#I = np.abs(rd*rd.conj()).real
#Ilm = cht.forward(I)
#phases = rd/np.sqrt(I)
#
#multiplicies = [1,0,1/100,1/10,1/2,2,10,100]
#intensities = []
#for m in multiplicies:
#    temp = [np.copy(Il)*m for Il in Ilm]
#    temp[0] = np.copy(Ilm[0])
#    intensities.append(cht.inverse(temp))
#
#db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/bl_from_cm/N_particles_shift.vtk',[d.copy() for d in intensities],grid=grid_pair.reciprocal,grid_type='spherical')

